#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateValueError
from libnmstate.schema import VLAN

from .base_iface import BaseIface


class VlanIface(BaseIface):
    @property
    def parent(self):
        return self._vlan_config.get(VLAN.BASE_IFACE)

    @property
    def need_parent(self):
        return True

    @property
    def _vlan_config(self):
        return self.raw.get(VLAN.CONFIG_SUBTREE, {})

    @property
    def is_virtual(self):
        return True

    @property
    def can_have_ip_as_port(self):
        return False

    def pre_edit_validation_and_cleanup(self):
        if self.is_up:
            self._validate_mandatory_properties()
        super().pre_edit_validation_and_cleanup()

    def _validate_mandatory_properties(self):
        for prop in (VLAN.ID, VLAN.BASE_IFACE):
            if prop not in self._vlan_config:
                raise NmstateValueError(
                    f"VLAN tunnel {self.name} has missing mandatory "
                    f"property: {prop}"
                )



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy
from operator import itemgetter

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.schema import OVSBridge
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType

from libnmstate.ifaces.ovs import OvsBridgeIface
from libnmstate.ifaces.ovs import OvsInternalIface
from libnmstate.ifaces.ifaces import Ifaces

from ..testlib.constants import PORT1_IFACE_NAME
from ..testlib.constants import PORT2_IFACE_NAME
from ..testlib.ifacelib import gen_foo_iface_info
from ..testlib.ovslib import OVS_BRIDGE_IFACE_NAME
from ..testlib.ovslib import OVS_IFACE_NAME
from ..testlib.ovslib import PORT_PORT_CONFIGS
from ..testlib.ovslib import gen_ovs_bridge_info


PORT_PORT_CONFIGS_VLAN_ACCESSS = [
    {
        OVSBridge.Port.NAME: PORT1_IFACE_NAME,
        OVSBridge.Port.VLAN_SUBTREE: {
            OVSBridge.Port.Vlan.MODE: OVSBridge.Port.Vlan.Mode.ACCESS,
            OVSBridge.Port.Vlan.TAG: 101,
        },
    },
    {
        OVSBridge.Port.NAME: PORT2_IFACE_NAME,
        OVSBridge.Port.VLAN_SUBTREE: {
            OVSBridge.Port.Vlan.MODE: OVSBridge.Port.Vlan.Mode.ACCESS,
            OVSBridge.Port.Vlan.TAG: 102,
        },
    },
]

BOND_IFACE_NAME = "bond0"

OVSBond = OVSBridge.Port.LinkAggregation

BOND_PORT_CONFIG = {
    OVSBridge.Port.NAME: BOND_IFACE_NAME,
    OVSBridge.Port.LINK_AGGREGATION_SUBTREE: {
        OVSBond.MODE: OVSBond.Mode.ACTIVE_BACKUP,
        OVSBond.PORT_SUBTREE: [
            {OVSBond.Port.NAME: PORT1_IFACE_NAME},
            {OVSBond.Port.NAME: PORT2_IFACE_NAME},
        ],
    },
}

PORT_IFACE_NAMES = sorted([s[OVSBridge.Port.NAME] for s in PORT_PORT_CONFIGS])


class TestOvsBrigeIface:
    def _gen_iface_info(self):
        iface_info = gen_foo_iface_info(iface_type=InterfaceType.OVS_BRIDGE)
        iface_info[OVSBridge.CONFIG_SUBTREE] = {
            OVSBridge.PORT_SUBTREE: deepcopy(PORT_PORT_CONFIGS),
            OVSBridge.OPTIONS_SUBTREE: {},
        }
        return iface_info

    def _gen_iface_info_with_bond(self):
        iface_info = gen_ovs_bridge_info()
        iface_info[OVSBridge.CONFIG_SUBTREE] = {
            OVSBridge.PORT_SUBTREE: [
                deepcopy(BOND_PORT_CONFIG),
                {OVSBridge.Port.NAME: OVS_IFACE_NAME},
            ],
            OVSBridge.OPTIONS_SUBTREE: {},
        }
        return iface_info

    def test_is_controller(self):
        iface = OvsBridgeIface(gen_ovs_bridge_info(), True)

        assert iface.is_controller

    def test_is_virtual(self):
        iface = OvsBridgeIface(gen_ovs_bridge_info(), True)

        assert iface.is_virtual

    def test_ovs_bridge_sort_port(self):
        iface1_info = gen_ovs_bridge_info()
        iface2_info = gen_ovs_bridge_info()
        iface2_info[OVSBridge.CONFIG_SUBTREE][OVSBridge.PORT_SUBTREE].reverse()

        iface1 = OvsBridgeIface(iface1_info, True)
        iface2 = OvsBridgeIface(iface2_info, True)

        assert iface1.state_for_verify() == iface2.state_for_verify()
        assert iface1.port == PORT_IFACE_NAMES
        assert iface2.port == PORT_IFACE_NAMES

    def test_ovs_bridge_sort_bond_port(self):
        iface1_info = self._gen_iface_info_with_bond()
        iface2_info = self._gen_iface_info_with_bond()
        bond_config = iface2_info[OVSBridge.CONFIG_SUBTREE][
            OVSBridge.PORT_SUBTREE
        ][0]
        bond_config[OVSBridge.Port.LINK_AGGREGATION_SUBTREE][
            OVSBond.PORT_SUBTREE
        ].reverse()

        iface1 = OvsBridgeIface(iface1_info, True)
        iface2 = OvsBridgeIface(iface2_info, True)

        assert iface1.state_for_verify() == iface2.state_for_verify()
        assert sorted(iface1.port) == PORT_IFACE_NAMES
        assert sorted(iface2.port) == PORT_IFACE_NAMES

    def test_gen_metadata_save_bond_port_config_to_port_iface(self):
        br_iface_info = self._gen_iface_info_with_bond()
        port1_iface_info = gen_foo_iface_info()
        port1_iface_info[Interface.NAME] = PORT1_IFACE_NAME
        port2_iface_info = gen_foo_iface_info()
        port2_iface_info[Interface.NAME] = PORT2_IFACE_NAME
        ifaces = Ifaces(
            des_iface_infos=[
                br_iface_info,
                port1_iface_info,
                port2_iface_info,
            ],
            cur_iface_infos=[port1_iface_info, port2_iface_info],
        )
        br_iface = ifaces.get_iface(
            OVS_BRIDGE_IFACE_NAME, InterfaceType.OVS_BRIDGE
        )
        br_iface.gen_metadata(ifaces)
        br_iface.pre_edit_validation_and_cleanup()
        port1_iface = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port2_iface = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]

        assert port1_iface.controller == OVS_BRIDGE_IFACE_NAME
        assert port2_iface.controller == OVS_BRIDGE_IFACE_NAME
        assert (
            port1_iface.to_dict()[OvsBridgeIface.BRPORT_OPTIONS_METADATA]
            == BOND_PORT_CONFIG
        )
        assert (
            port2_iface.to_dict()[OvsBridgeIface.BRPORT_OPTIONS_METADATA]
            == BOND_PORT_CONFIG
        )

    def test_auto_create_ovs_interface(self):
        iface_info = gen_ovs_bridge_info()
        port1_iface_info = gen_foo_iface_info()
        port1_iface_info[Interface.NAME] = PORT1_IFACE_NAME
        port2_iface_info = gen_foo_iface_info()
        port2_iface_info[Interface.NAME] = PORT2_IFACE_NAME
        ifaces = Ifaces(
            des_iface_infos=[iface_info, port1_iface_info, port2_iface_info],
            cur_iface_infos=[port1_iface_info, port2_iface_info],
        )

        ovs_iface = ifaces.all_kernel_ifaces[OVS_IFACE_NAME]
        assert ovs_iface.type == InterfaceType.OVS_INTERFACE
        assert ovs_iface.parent == OVS_BRIDGE_IFACE_NAME
        assert ovs_iface.is_virtual
        assert ovs_iface.controller == OVS_BRIDGE_IFACE_NAME

    def test_validate_ovs_bond_with_single_port(self):
        iface_info = self._gen_iface_info_with_bond()
        bond_config = iface_info[OVSBridge.CONFIG_SUBTREE][
            OVSBridge.PORT_SUBTREE
        ][0][OVSBridge.Port.LINK_AGGREGATION_SUBTREE]
        bond_config[OVSBond.PORT_SUBTREE].pop()

        iface = OvsBridgeIface(iface_info, True)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_remove_port_from_bridge_without_bond(self):
        iface = OvsBridgeIface(gen_ovs_bridge_info(), True)
        iface.remove_port(PORT1_IFACE_NAME)

        assert sorted(iface.port) == sorted([PORT2_IFACE_NAME, OVS_IFACE_NAME])
        assert iface.to_dict()[OVSBridge.CONFIG_SUBTREE][
            OVSBridge.PORT_SUBTREE
        ] == (
            sorted(
                [
                    {OVSBridge.Port.NAME: PORT2_IFACE_NAME},
                    {OVSBridge.Port.NAME: OVS_IFACE_NAME},
                ],
                key=itemgetter(OVSBridge.Port.NAME),
            )
        )

    def test_remove_port_from_bridge_with_bond(self):
        iface = OvsBridgeIface(self._gen_iface_info_with_bond(), True)
        iface.remove_port(PORT1_IFACE_NAME)
        expected_port_config = deepcopy(BOND_PORT_CONFIG)
        expected_port_config[OVSBridge.Port.LINK_AGGREGATION_SUBTREE][
            OVSBond.PORT_SUBTREE
        ].pop(0)

        assert sorted(iface.port) == sorted([PORT2_IFACE_NAME, OVS_IFACE_NAME])
        assert iface.to_dict()[OVSBridge.CONFIG_SUBTREE][
            OVSBridge.PORT_SUBTREE
        ] == sorted(
            [{OVSBridge.Port.NAME: OVS_IFACE_NAME}, expected_port_config],
            key=itemgetter(OVSBridge.Port.NAME),
        )


class TestOvsInternalIface:
    def _gen_iface_info(self):
        return gen_foo_iface_info(iface_type=InterfaceType.OVS_INTERFACE)

    def test_is_virtual(self):
        assert OvsInternalIface(gen_ovs_bridge_info()).is_virtual

    def test_need_parent(self):
        assert OvsInternalIface(gen_ovs_bridge_info()).need_parent

    def test_can_have_ip_as_port(self):
        assert OvsInternalIface(gen_ovs_bridge_info()).can_have_ip_as_port

    # The 'parent' property is tested by `test_auto_create_ovs_interface`.



from setuptools import setup, find_packages
from datetime import date


def readme():
    with open("README.md") as f:
        return f.read()


def requirements():
    req = []
    with open("requirements.txt") as fd:
        for line in fd:
            line.strip()
            if not line.startswith("#"):
                req.append(line)
    return req


def get_version():
    with open("libnmstate/VERSION") as f:
        version = f.read().strip()
    return version


def gen_manpage():
    manpage = ""
    with open("doc/nmstatectl.8.in") as f:
        manpage = f.read()
    manpage = manpage.replace("@DATE@", date.today().strftime("%B %d, %Y"))
    manpage = manpage.replace("@VERSION@", get_version())
    with open("doc/nmstatectl.8", "w") as f:
        f.write(manpage)
    return [("share/man/man8", ["doc/nmstatectl.8"])]


setup(
    name="nmstate",
    version=get_version(),
    description="Declarative network manager API",
    author="Edward Haas",
    author_email="ehaas@redhat.com",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://nmstate.github.io/",
    license="LGPL2.1+",
    packages=find_packages(),
    install_requires=requirements(),
    entry_points={
        "console_scripts": ["nmstatectl = nmstatectl.nmstatectl:main"]
    },
    package_data={
        "libnmstate": ["schemas/operational-state.yaml", "VERSION"],
        "nmstatectl": ["io.nmstate.varlink", "nmstate-varlink.service"],
    },
    data_files=gen_manpage(),
)



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import os

import gi

gi.require_version("NM", "1.0")
# It is required to state the NM version before importing it
# But this break the flak8 rule: https://www.flake8rules.com/rules/E402.html
# Use NOQA: E402 to suppress it.
from gi.repository import NM  # NOQA: E402


def is_fedora():
    return os.path.exists("/etc/fedora-release")


def is_ubuntu_kernel():
    return "Ubuntu" in os.uname().version


def nm_major_minor_version():
    return float(f"{NM.MAJOR_VERSION}.{NM.MINOR_VERSION}")



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging

from libnmstate.error import NmstateLibnmError
from libnmstate.error import NmstateInternalError

from .common import GLib
from .common import NM
from .device import get_nm_dev
from .device import get_iface_type
from .device import mark_device_as_managed
from .ipv4 import is_dynamic as is_ipv4_dynamic
from .ipv6 import is_dynamic as is_ipv6_dynamic


NM_AC_STATE_CHANGED_SIGNAL = "state-changed"


def is_activated(nm_ac, nm_dev):
    if not (nm_ac and nm_dev):
        return False

    state = nm_ac.get_state()
    if state == NM.ActiveConnectionState.ACTIVATED:
        return True
    elif state == NM.ActiveConnectionState.ACTIVATING:
        ac_state_flags = nm_ac.get_state_flags()
        nm_flags = NM.ActivationStateFlags
        ip4_is_dynamic = is_ipv4_dynamic(nm_ac)
        ip6_is_dynamic = is_ipv6_dynamic(nm_ac)
        if (
            ac_state_flags & nm_flags.IS_MASTER
            or (ip4_is_dynamic and ac_state_flags & nm_flags.IP6_READY)
            or (ip6_is_dynamic and ac_state_flags & nm_flags.IP4_READY)
            or (ip4_is_dynamic and ip6_is_dynamic)
        ):
            # For interface meet any condition below will be
            # treated as activated when reach IP_CONFIG state:
            #   * Is controller device.
            #   * DHCPv4 enabled with IP6_READY flag.
            #   * DHCPv6/Autoconf with IP4_READY flag.
            #   * DHCPv4 enabled with DHCPv6/Autoconf enabled.
            return (
                NM.DeviceState.IP_CONFIG
                <= nm_dev.get_state()
                <= NM.DeviceState.ACTIVATED
            )

    return False


def is_activating(nm_ac, nm_dev):
    if not nm_ac or not nm_dev:
        return True
    if nm_dev.get_state_reason() == NM.DeviceStateReason.NEW_ACTIVATION:
        return True

    return (
        nm_ac.get_state() == NM.ActiveConnectionState.ACTIVATING
    ) and not is_activated(nm_ac, nm_dev)


class ProfileActivation:
    def __init__(self, ctx, iface_name, iface_type, nm_profile, nm_dev):
        self._ctx = ctx
        self._iface_name = iface_name
        self._iface_type = iface_type
        self._nm_ac = None
        self._nm_dev = nm_dev
        self._nm_profile = nm_profile
        self._ac_handlers = set()
        self._dev_handlers = set()
        self._action = None

    def run(self):
        specific_object = None
        self._action = (
            f"Activate profile uuid:{self._nm_profile.get_uuid()} "
            f"iface:{self._iface_name} type: {self._iface_type}"
        )
        if self._nm_dev:
            # Workaround of https://bugzilla.redhat.com/1880420
            mark_device_as_managed(self._ctx, self._nm_dev)

        user_data = None
        self._ctx.register_async(self._action)
        self._ctx.client.activate_connection_async(
            self._nm_profile,
            self._nm_dev,
            specific_object,
            self._ctx.cancellable,
            self._activate_profile_callback,
            user_data,
        )

    @staticmethod
    def wait(ctx, nm_ac, nm_dev):
        activation = ProfileActivation(
            ctx,
            nm_dev.get_iface(),
            get_iface_type(nm_dev),
            None,
            nm_dev,
        )
        activation._nm_ac = nm_ac
        activation._action = (
            f"Waiting activation of {activation._iface_name} "
            f"{activation._iface_type}"
        )
        ctx.register_async(activation._action)
        activation._wait_profile_activation()

    def _activate_profile_callback(self, nm_client, result, _user_data):
        nm_ac = None
        if self._ctx.is_cancelled():
            self._activation_clean_up()
            return
        try:
            nm_ac = nm_client.activate_connection_finish(result)
        except Exception as e:
            self._ctx.fail(
                NmstateLibnmError(f"{self._action} failed: error={e}")
            )

        if nm_ac is None:
            self._ctx.fail(
                NmstateLibnmError(
                    f"{self._action} failed: "
                    "error='None return from activate_connection_finish()'"
                )
            )
        else:
            logging.debug(
                f"Connection activation initiated: iface={self._iface_name} "
                f"type={self._iface_type} con-state={nm_ac.get_state()}"
            )
            self._nm_ac = nm_ac
            self._nm_dev = get_nm_dev(
                self._ctx, self._iface_name, self._iface_type
            )
            self._wait_profile_activation()

    def _wait_profile_activation(self):
        if is_activated(self._nm_ac, self._nm_dev):
            logging.debug(
                "Connection activation succeeded: "
                f"iface={self._iface_name}, type={self._iface_type}, "
                f"con_state={self._nm_ac.get_state()}, "
                f"dev_state={self._nm_dev.get_state()}, "
                f"state_flags={self._nm_ac.get_state_flags()}"
            )
            self._activation_clean_up()
            self._ctx.finish_async(self._action)
        elif is_activating(self._nm_ac, self._nm_dev):
            if self._nm_ac:
                self._wait_nm_ac_activation()
            if self._nm_dev:
                self._wait_nm_dev_activation()
            if not self._nm_ac and not self._nm_dev:
                self._ctx.fail(
                    NmstateInternalError(
                        f"{self._action} failed: no nm_ac or nm_dev"
                    )
                )
        else:
            if self._nm_dev:
                error_msg = (
                    f"Connection {self._nm_profile.get_uuid()} failed: "
                    f"state={self._nm_ac.get_state()} "
                    f"reason={self._nm_ac.get_state_reason()} "
                    f"dev_state={self._nm_dev.get_state()} "
                    f"dev_reason={self._nm_dev.get_state_reason()}"
                )
            else:
                error_msg = (
                    f"Connection {self._nm_profile.get_uuid()} failed: "
                    f"state={self._nm_ac.get_state()} "
                    f"reason={self._nm_ac.get_state_reason()} dev=None"
                )
            self._activation_clean_up()
            logging.error(error_msg)
            self._ctx.fail(
                NmstateLibnmError(f"{self._action} failed: {error_msg}")
            )

    def _activation_clean_up(self):
        self._remove_ac_handlers()
        self._remove_dev_handlers()

    def _remove_ac_handlers(self):
        for handler_id in self._ac_handlers:
            self._nm_ac.handler_disconnect(handler_id)
        self._ac_handlers = set()

    def _remove_dev_handlers(self):
        for handler_id in self._dev_handlers:
            self._nm_dev.handler_disconnect(handler_id)
        self._dev_handlers = set()

    def _wait_nm_ac_activation(self):
        user_data = None
        self._ac_handlers.add(
            self._nm_ac.connect(
                NM_AC_STATE_CHANGED_SIGNAL,
                self._ac_state_change_callback,
                user_data,
            )
        )
        self._ac_handlers.add(
            self._nm_ac.connect(
                "notify::state-flags",
                self._ac_state_flags_change_callback,
                user_data,
            )
        )

    def _ac_state_change_callback(self, _nm_ac, _state, _reason, _user_data):
        if self._ctx.is_cancelled():
            self._activation_clean_up()
            return
        self._activation_progress_check()

    def _ac_state_flags_change_callback(self, _nm_ac, _state, _user_data):
        if self._ctx.is_cancelled():
            self._activation_clean_up()
            return
        self._activation_progress_check()

    def _wait_nm_dev_activation(self):
        user_data = None
        self._dev_handlers.add(
            self._nm_dev.connect(
                "state-changed", self._dev_state_change_callback, user_data
            )
        )

    def _dev_state_change_callback(
        self, _nm_dev, _new_state, _old_state, _reason, _user_data
    ):
        if self._ctx.is_cancelled():
            self._activation_clean_up()
            return
        self._activation_progress_check()

    def _activation_progress_check(self):
        cur_nm_dev = get_nm_dev(self._ctx, self._iface_name, self._iface_type)
        if cur_nm_dev and cur_nm_dev != self._nm_dev:
            logging.debug(
                f"The NM.Device of profile {self._iface_name} "
                f"{self._iface_type} changed"
            )
            self._remove_dev_handlers()
            self._nm_dev = cur_nm_dev
            self._wait_nm_dev_activation()

        if cur_nm_dev:
            cur_nm_ac = cur_nm_dev.get_active_connection()
            if cur_nm_ac and cur_nm_ac != self._nm_ac:
                logging.debug(
                    f"Active connection of device {self._iface_name} "
                    "has been replaced"
                )
                self._remove_ac_handlers()
                self._nm_ac = cur_nm_ac
                self._wait_nm_ac_activation()

        if is_activated(self._nm_ac, self._nm_dev):
            logging.debug(
                "Connection activation succeeded: "
                f"iface={self._iface_name}, type={self._iface_type}, "
                f"con_state={self._nm_ac.get_state()}, "
                f"dev_state={self._nm_dev.get_state()}, "
                f"state_flags={self._nm_ac.get_state_flags()}"
            )
            self._activation_clean_up()
            self._ctx.finish_async(self._action)
        elif not is_activating(self._nm_ac, self._nm_dev):
            reason = f"{self._nm_ac.get_state_reason()}"
            if self._nm_dev:
                reason += f" {self._nm_dev.get_state_reason()}"
            self._activation_clean_up()
            self._ctx.fail(
                NmstateLibnmError(f"{self._action} failed: reason={reason}")
            )


class ActiveConnectionDeactivate:
    def __init__(self, ctx, iface_name, iface_type, nm_ac):
        self._ctx = ctx
        self._iface_name = iface_name
        self._iface_type = iface_type
        self._nm_ac = nm_ac

    def run(self):
        if self._nm_ac.props.state == NM.ActiveConnectionState.DEACTIVATED:
            return

        action = f"Deactivate profile: {self._iface_name} {self._iface_type}"
        self._ctx.register_async(action)
        handler_id = self._nm_ac.connect(
            NM_AC_STATE_CHANGED_SIGNAL,
            self._wait_state_changed_callback,
            action,
        )
        if self._nm_ac.props.state != NM.ActiveConnectionState.DEACTIVATING:
            user_data = (handler_id, action)
            self._ctx.client.deactivate_connection_async(
                self._nm_ac,
                self._ctx.cancellable,
                self._deactivate_connection_callback,
                user_data,
            )

    def _wait_state_changed_callback(self, nm_ac, state, reason, action):
        if self._ctx.is_cancelled():
            return
        if nm_ac.props.state == NM.ActiveConnectionState.DEACTIVATED:
            logging.debug(
                "Connection deactivation succeeded on %s",
                self._iface_name,
            )
            self._ctx.finish_async(action)

    def _deactivate_connection_callback(self, nm_client, result, user_data):
        handler_id, action = user_data
        if self._ctx.is_cancelled():
            if self._nm_ac:
                self._nm_ac.handler_disconnect(handler_id)
            return

        try:
            success = nm_client.deactivate_connection_finish(result)
        except GLib.Error as e:
            if e.matches(
                NM.ManagerError.quark(), NM.ManagerError.CONNECTIONNOTACTIVE
            ):
                success = True
                logging.debug(
                    "Connection is not active on {}, no need to "
                    "deactivate".format(self._iface_name)
                )
                if self._nm_ac:
                    self._nm_ac.handler_disconnect(handler_id)
                self._ctx.finish_async(action)
            else:
                if self._nm_ac:
                    self._nm_ac.handler_disconnect(handler_id)
                self._ctx.fail(
                    NmstateLibnmError(f"{action} failed: error={e}")
                )
                return
        except Exception as e:
            if self._nm_ac:
                self._nm_ac.handler_disconnect(handler_id)
            self._ctx.fail(
                NmstateLibnmError(
                    "BUG: Unexpected error when activating "
                    f"{self._iface_name} error={e}"
                )
            )
            return

        if not success:
            if self._nm_ac:
                self._nm_ac.handler_disconnect(handler_id)
            self._ctx.fail(
                NmstateLibnmError(
                    f"{action} failed: error='None returned from "
                    "deactivate_connection_finish()'"
                )
            )



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging
import subprocess
import warnings

import pytest

import libnmstate
from libnmstate.schema import DNS
from libnmstate.schema import Route
from libnmstate.schema import RouteRule

from .testlib import ifacelib


REPORT_HEADER = """RPMs: {rpms}
OS: {osname}
nmstate: {nmstate_version}
"""


def pytest_configure(config):
    config.addinivalue_line("markers", "slow tier1 tier2")


def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--runslow"):
        # --runslow is not in cli: skip slow tests
        _mark_skip_slow_tests(items)
    _mark_tier2_tests(items)


def _mark_skip_slow_tests(items):
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


def _mark_tier2_tests(items):
    for item in items:
        if "tier1" not in item.keywords:
            item.add_marker(pytest.mark.tier2)


@pytest.fixture(scope="session", autouse=True)
def test_env_setup():
    _logging_setup()
    old_state = libnmstate.show()
    _empty_net_state()
    _ethx_init()
    yield
    libnmstate.apply(old_state, verify_change=False)
    _diff_initial_state(old_state)


def _empty_net_state():
    """
    Remove existing DNS, routes, route rules in case it interference tests.
    """
    desired_state = libnmstate.show()
    desired_state[DNS.KEY] = {DNS.CONFIG: {}}
    desired_state[Route.KEY] = {
        Route.CONFIG: [{Route.STATE: Route.STATE_ABSENT}]
    }
    desired_state[RouteRule.KEY] = {RouteRule.CONFIG: []}

    libnmstate.apply(desired_state)


def _logging_setup():
    logging.basicConfig(
        format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
        level=logging.DEBUG,
    )


def _ethx_init():
    """ Remove any existing definitions on the ethX interfaces. """
    ifacelib.ifaces_init("eth1", "eth2")


@pytest.fixture(scope="function")
def eth1_up():
    with ifacelib.iface_up("eth1") as ifstate:
        yield ifstate


@pytest.fixture(scope="function")
def eth2_up():
    with ifacelib.iface_up("eth2") as ifstate:
        yield ifstate


port0_up = eth1_up
port1_up = eth2_up


def _diff_initial_state(old_state):
    new_state = libnmstate.show()

    if old_state != new_state:
        warnings.warn(
            "Network state after test run does not match network state "
            "before test run:\n {}\n".format(
                libnmstate.prettystate.format_desired_current_state_diff(
                    old_state, new_state
                )
            )
        )


def pytest_report_header(config):
    return REPORT_HEADER.format(
        rpms=_get_package_nvr("NetworkManager"),
        osname=_get_osname(),
        nmstate_version=_get_nmstate_version(),
    )


def _get_nmstate_version():
    """
    Prefer RPM version of nmstate, if not found, use libnmstate module version
    """
    try:
        return _get_package_nvr("nmstate")
    except subprocess.CalledProcessError:
        return libnmstate.__version__


def _get_package_nvr(package):
    return (
        subprocess.check_output(["rpm", "-q", package]).strip().decode("utf-8")
    )


def _get_osname():
    with open("/etc/os-release") as os_release:
        for line in os_release.readlines():
            if line.startswith("PRETTY_NAME="):
                return line.split("=", maxsplit=1)[1].strip().strip('"')
    return ""



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from contextlib import contextmanager
import time

import pytest

import libnmstate
from libnmstate.error import NmstateVerificationError

from .testlib import assertlib
from .testlib import cmdlib
from .testlib import statelib
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState

DUMMY_INTERFACE = "dummy_test"


@pytest.fixture(scope="function")
def ip_link_dummy():
    cmdlib.exec_cmd(["ip", "link", "add", DUMMY_INTERFACE, "type", "dummy"])
    try:
        yield
    finally:
        cmdlib.exec_cmd(["ip", "link", "del", DUMMY_INTERFACE])


@contextmanager
def dummy_interface(name):
    dummy_desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: name,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {InterfaceIPv4.ENABLED: False},
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
            }
        ]
    }
    libnmstate.apply(dummy_desired_state)
    try:
        yield dummy_desired_state
    finally:
        dummy_state = dummy_desired_state[Interface.KEY][0]
        dummy_state[Interface.STATE] = InterfaceState.ABSENT
        libnmstate.apply(dummy_desired_state)


def test_iface_description_removal(eth1_up):
    desired_state = eth1_up
    desired_state[Interface.KEY][0][Interface.DESCRIPTION] = "bar"
    libnmstate.apply(desired_state)
    current_state = statelib.show_only(("eth1",))
    assert current_state[Interface.KEY][0][Interface.DESCRIPTION] == "bar"

    desired_state[Interface.KEY][0][Interface.DESCRIPTION] = ""
    libnmstate.apply(desired_state)
    current_state = statelib.show_only(("eth1",))
    assert Interface.DESCRIPTION not in current_state[Interface.KEY][0]


def test_take_over_virtual_interface_then_remove(ip_link_dummy):
    with dummy_interface(DUMMY_INTERFACE) as dummy_desired_state:
        assertlib.assert_state_match(dummy_desired_state)

    current_state = statelib.show_only((DUMMY_INTERFACE,))
    assert len(current_state[Interface.KEY]) == 0


def test_take_over_virtual_interface_and_rollback(ip_link_dummy):
    with dummy_interface(DUMMY_INTERFACE) as dummy_desired_state:
        assertlib.assert_state_match(dummy_desired_state)

        dummy_desired_state[Interface.KEY][0]["invalid_key"] = "foo"
        with pytest.raises(NmstateVerificationError):
            libnmstate.apply(dummy_desired_state)

        time.sleep(5)

        current_state = statelib.show_only((DUMMY_INTERFACE,))
        assert len(current_state[Interface.KEY]) == 1



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

import libnmstate

from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from .testlib import assertlib


def test_set_a_down_iface_down(eth1_up):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.DOWN,
            }
        ]
    }
    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.xfail(
    raises=AssertionError,
    reason="Some ifaces cannot be removed",
    strict=True,
)
def test_removing_a_non_removable_iface(eth1_up):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.ABSENT,
            }
        ]
    }

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


def test_set_iface_down_without_type(eth1_up):
    desired_state = {
        Interface.KEY: [
            {Interface.NAME: "eth1", Interface.STATE: InterfaceState.DOWN}
        ]
    }
    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


def test_change_iface_without_type(eth1_up):
    desired_state = {
        Interface.KEY: [{Interface.NAME: "eth1", Interface.MTU: 1400}]
    }
    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import Route

from libnmstate.ifaces import BaseIface
from libnmstate.route import RouteEntry
from libnmstate.route import RouteState

from .testlib.ifacelib import gen_two_static_ip_ifaces
from .testlib.routelib import IPV4_ROUTE_IFACE_NAME
from .testlib.routelib import IPV4_ROUTE_DESITNATION
from .testlib.routelib import IPV6_ROUTE_IFACE_NAME
from .testlib.routelib import gen_ipv4_route
from .testlib.routelib import gen_ipv6_route


parametrize_route_property = pytest.mark.parametrize(
    "route_property",
    [
        Route.TABLE_ID,
        Route.DESTINATION,
        Route.NEXT_HOP_INTERFACE,
        Route.NEXT_HOP_ADDRESS,
        Route.METRIC,
    ],
)


class TestRouteEntry:
    def test_hash_unique(self):
        route = _create_route("198.51.100.0/24", "192.0.2.1", "eth1", 50, 103)
        assert hash(route) == hash(route)

    def test_obj_unique(self):
        route0 = _create_route("198.51.100.0/24", "192.0.2.1", "eth1", 50, 103)
        route1 = _create_route(
            "2001:db8:a::/64", "2001:db8:1::a", "eth2", 51, 104
        )
        route0_clone = _create_route(
            "198.51.100.0/24", "192.0.2.1", "eth1", 50, 103
        )
        assert route0 == route0_clone
        assert route0 != route1

    def test_obj_unique_without_table_id(self):
        route_with_default_table_id = _create_route(
            "198.51.100.0/24",
            "192.0.2.1",
            "eth1",
            Route.USE_DEFAULT_ROUTE_TABLE,
            103,
        )

        route_without_table_id = _create_route(
            "198.51.100.0/24", "192.0.2.1", "eth1", None, 103
        )

        assert route_without_table_id == route_with_default_table_id

    def test_obj_unique_without_metric(self):
        route_with_default_metric = _create_route(
            "198.51.100.0/24",
            "192.0.2.1",
            "eth1",
            50,
            Route.USE_DEFAULT_METRIC,
        )

        route_without_metric = _create_route(
            "198.51.100.0/24", "192.0.2.1", "eth1", 50, None
        )

        assert route_without_metric == route_with_default_metric

    def test_obj_unique_without_next_hop(self):
        route_with_default_next_hop = _create_route(
            "198.51.100.0/24", "", "eth1", 50, 103
        )

        route_without_next_hop = _create_route(
            "198.51.100.0/24", None, "eth1", 50, 103
        )

        assert route_without_next_hop == route_with_default_next_hop

    def test_normal_route_object_as_dict(self):
        route = _create_route_dict(
            "198.51.100.0/24", "192.0.2.1", "eth1", 50, 103
        )
        route_obj = RouteEntry(route)
        assert route_obj.to_dict() == route

    def test_absent_route_object_as_dict(self):
        route = _create_route_dict(
            "198.51.100.0/24", "192.0.2.1", "eth1", 50, 103
        )
        route[Route.STATE] = Route.STATE_ABSENT
        route_obj = RouteEntry(route)
        assert route_obj.absent
        assert route_obj.to_dict() == route

    @parametrize_route_property
    def test_absent_route_with_missing_props_as_dict(self, route_property):
        absent_route = _create_route_dict(
            "198.51.100.0/24", "192.0.2.1", "eth1", 50, 103
        )
        absent_route[Route.STATE] = Route.STATE_ABSENT
        del absent_route[route_property]
        route_obj = RouteEntry(absent_route)
        assert route_obj.to_dict() == absent_route

    def test_absent_route_with_exact_match(self):
        route0 = _create_route("198.51.100.0/24", "192.0.2.1", "eth1", 50, 103)

        absent_r0 = _create_route_dict(
            "198.51.100.0/24", "192.0.2.1", "eth1", 50, 103
        )
        absent_r0[Route.STATE] = Route.STATE_ABSENT
        absent_route0 = RouteEntry(absent_r0)

        route1 = _create_route(
            "2001:db8:a::/64", "2001:db8:1::a", "eth2", 51, 104
        )

        assert absent_route0.match(route0)
        assert absent_route0 == route0
        assert not absent_route0.match(route1)
        assert absent_route0 != route1

    @parametrize_route_property
    def test_absent_route_wildcard_match(self, route_property):
        original_route0 = _create_route(
            "198.51.100.0/24", "192.0.2.1", "eth1", 50, 103
        )
        original_route1 = _create_route(
            "2001:db8:a::/64", "2001:db8:1::a", "eth2", 51, 104
        )

        absent_route0_state = _create_route_dict(
            "198.51.100.0/24", "192.0.2.1", "eth1", 50, 103
        )
        absent_route0_state[Route.STATE] = Route.STATE_ABSENT
        del absent_route0_state[route_property]
        new_route0 = RouteEntry(absent_route0_state)

        assert new_route0.match(original_route0)
        assert not new_route0.match(original_route1)

    def test_absent_route_is_ignored_for_matching_and_equality(self):
        route = _create_route_dict(
            "198.51.100.0/24", "192.0.2.1", "eth1", 50, 103
        )
        route[Route.STATE] = Route.STATE_ABSENT
        obj1 = RouteEntry(route)
        obj2 = RouteEntry(route)
        assert obj1.match(obj2)
        assert obj1 == obj2

    def test_sort_routes(self):
        routes = [
            _create_route("198.51.100.1/24", "192.0.2.1", "eth0", 50, 103),
            _create_route("198.51.100.0/24", "192.0.2.1", "eth0", 50, 103),
            _create_route("198.51.100.0/24", "192.0.2.3", "eth1", 10, 103),
            _create_route("198.51.100.0/24", "192.0.2.1", "eth1", 10, 101),
        ]
        expected_routes = [
            _create_route("198.51.100.0/24", "192.0.2.1", "eth1", 10, 101),
            _create_route("198.51.100.0/24", "192.0.2.3", "eth1", 10, 103),
            _create_route("198.51.100.0/24", "192.0.2.1", "eth0", 50, 103),
            _create_route("198.51.100.1/24", "192.0.2.1", "eth0", 50, 103),
        ]
        assert expected_routes == sorted(routes)

    @parametrize_route_property
    def test_sort_routes_with_absent_route(self, route_property):
        absent_route = _create_route(
            "198.51.100.0/24", "192.0.1.1", "eth0", 9, 103
        ).to_dict()
        absent_route[Route.STATE] = Route.STATE_ABSENT
        del absent_route[route_property]
        absent_route = RouteEntry(absent_route)
        routes = [
            absent_route,
            _create_route("198.51.100.1/24", "192.0.2.1", "eth0", 50, 103),
            _create_route("198.51.100.0/24", "192.0.2.1", "eth0", 50, 103),
            _create_route("198.51.100.0/24", "192.0.2.1", "eth1", 10, 103),
        ]
        expected_routes = [
            absent_route,
            _create_route("198.51.100.0/24", "192.0.2.1", "eth1", 10, 103),
            _create_route("198.51.100.0/24", "192.0.2.1", "eth0", 50, 103),
            _create_route("198.51.100.1/24", "192.0.2.1", "eth0", 50, 103),
        ]
        assert expected_routes == sorted(routes)


class TestRouteState:
    def _gen_ifaces(self):
        return gen_two_static_ip_ifaces(
            IPV4_ROUTE_IFACE_NAME, IPV6_ROUTE_IFACE_NAME
        )

    def _gen_route_state(self, des_routes, cur_routes):
        return RouteState(
            self._gen_ifaces(),
            {Route.CONFIG: [r.to_dict() for r in des_routes]},
            {Route.CONFIG: [r.to_dict() for r in cur_routes]},
        )

    def test_merge_empty_states(self):
        state = RouteState(self._gen_ifaces(), [], [])

        assert {} == state.config_iface_routes

    def test_merge_identical_states(self):
        ipv4_route = gen_ipv4_route()

        state = self._gen_route_state([ipv4_route], [ipv4_route])

        assert [IPV4_ROUTE_IFACE_NAME] == list(
            state.config_iface_routes.keys()
        )
        assert state.config_iface_routes[IPV4_ROUTE_IFACE_NAME] == set(
            [ipv4_route]
        )

    def test_merge_unique_states(self):
        ipv4_route = gen_ipv4_route()
        ipv6_route = gen_ipv6_route()
        state = self._gen_route_state([ipv4_route], [ipv6_route])

        assert [IPV4_ROUTE_IFACE_NAME, IPV6_ROUTE_IFACE_NAME] == sorted(
            list(state.config_iface_routes.keys())
        )
        assert state.config_iface_routes[IPV4_ROUTE_IFACE_NAME] == set(
            [ipv4_route]
        )
        assert state.config_iface_routes[IPV6_ROUTE_IFACE_NAME] == set(
            [ipv6_route]
        )

    def test_merge_empty_with_non_empty_state(self):
        ipv4_route = gen_ipv4_route()

        state = self._gen_route_state([], [ipv4_route])
        assert [IPV4_ROUTE_IFACE_NAME] == list(
            state.config_iface_routes.keys()
        )
        assert state.config_iface_routes[IPV4_ROUTE_IFACE_NAME] == set(
            [ipv4_route]
        )

    def test_remove_route_with_exact_match(self):
        ipv4_route = gen_ipv4_route()
        ipv4_route_absent_dict = gen_ipv4_route().to_dict()
        ipv4_route_absent_dict[Route.STATE] = Route.STATE_ABSENT
        ipv4_route_absent = RouteEntry(ipv4_route_absent_dict)

        state = self._gen_route_state([ipv4_route_absent], [ipv4_route])

        assert not list(state.config_iface_routes.keys())

    def test_remove_route_with_wildcard_match_with_next_hop_iface(self):
        ipv4_route = gen_ipv4_route()
        route_absent = RouteEntry(
            {
                Route.NEXT_HOP_INTERFACE: IPV4_ROUTE_IFACE_NAME,
                Route.STATE: Route.STATE_ABSENT,
            }
        )
        state = self._gen_route_state([route_absent], [ipv4_route])
        assert not list(state.config_iface_routes.keys())

    def test_remove_route_with_wildcard_match_without_next_hop_iface(self):
        ipv4_route = gen_ipv4_route()
        route_absent = RouteEntry(
            {
                Route.DESTINATION: IPV4_ROUTE_DESITNATION,
                Route.STATE: Route.STATE_ABSENT,
            }
        )
        state = self._gen_route_state([route_absent], [ipv4_route])
        assert not list(state.config_iface_routes.keys())

    def test_validate_route_to_abent_iface(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[
            IPV4_ROUTE_IFACE_NAME
        ].state = InterfaceState.ABSENT

        ipv4_route = gen_ipv4_route()
        with pytest.raises(NmstateValueError):
            RouteState(ifaces, {Route.CONFIG: [ipv4_route.to_dict()]}, [])

    def test_validate_route_to_down_iface(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[
            IPV4_ROUTE_IFACE_NAME
        ].state = InterfaceState.DOWN

        ipv4_route = gen_ipv4_route()
        with pytest.raises(NmstateValueError):
            RouteState(ifaces, {Route.CONFIG: [ipv4_route.to_dict()]}, [])

    def test_validate_route_to_ipv4_disabled(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME].raw[Interface.IPV4][
            InterfaceIPv4.ENABLED
        ] = False

        ipv4_route = gen_ipv4_route()
        with pytest.raises(NmstateValueError):
            RouteState(ifaces, {Route.CONFIG: [ipv4_route.to_dict()]}, [])

    def test_validate_route_to_ipv6_disabled(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME].raw[Interface.IPV6][
            InterfaceIPv6.ENABLED
        ] = False

        ipv6_route = gen_ipv6_route()
        with pytest.raises(NmstateValueError):
            RouteState(ifaces, {Route.CONFIG: [ipv6_route.to_dict()]}, [])

    def test_validate_route_without_next_hop_iface(self):
        ipv4_route_dict = gen_ipv4_route().to_dict()
        ipv4_route_dict.pop(Route.NEXT_HOP_INTERFACE)
        with pytest.raises(NmstateValueError):
            self._gen_route_state([RouteEntry(ipv4_route_dict)], [])

    def test_validate_route_without_destination(self):
        ipv4_route_dict = gen_ipv4_route().to_dict()
        ipv4_route_dict.pop(Route.DESTINATION)
        with pytest.raises(NmstateValueError):
            self._gen_route_state([RouteEntry(ipv4_route_dict)], [])

    def test_validate_route_next_hop_to_unknown_iface(self):
        ipv4_route_dict = gen_ipv4_route().to_dict()
        ipv4_route_dict[Route.NEXT_HOP_INTERFACE] = "not_exit_iface"
        with pytest.raises(NmstateValueError):
            self._gen_route_state([RouteEntry(ipv4_route_dict)], [])

    def test_validate_route_next_hop_to_dhcpv4_ifce(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME].raw[Interface.IPV4][
            InterfaceIPv4.DHCP
        ] = True

        ipv4_route = gen_ipv4_route()
        with pytest.raises(NmstateValueError):
            RouteState(ifaces, {Route.CONFIG: [ipv4_route.to_dict()]}, [])

    def test_validate_route_next_hop_to_dhcpv6_ifce(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME].raw[Interface.IPV6][
            InterfaceIPv6.DHCP
        ] = True

        ipv6_route = gen_ipv6_route()
        with pytest.raises(NmstateValueError):
            RouteState(ifaces, {Route.CONFIG: [ipv6_route.to_dict()]}, [])

    def test_validate_route_next_hop_to_ipv6_autoconf_ifce(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME].raw[Interface.IPV6][
            InterfaceIPv6.AUTOCONF
        ] = True

        ipv6_route = gen_ipv6_route()
        with pytest.raises(NmstateValueError):
            RouteState(ifaces, {Route.CONFIG: [ipv6_route.to_dict()]}, [])

    def test_discard_route_next_hop_absent_iface(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[
            IPV4_ROUTE_IFACE_NAME
        ].state = InterfaceState.ABSENT

        ipv4_route = gen_ipv4_route()
        ipv6_route = gen_ipv6_route()
        state = RouteState(
            ifaces,
            [],
            {Route.CONFIG: [ipv4_route.to_dict(), ipv6_route.to_dict()]},
        )
        assert [IPV6_ROUTE_IFACE_NAME] == list(
            state.config_iface_routes.keys()
        )
        assert state.config_iface_routes[IPV6_ROUTE_IFACE_NAME] == set(
            [ipv6_route]
        )

    def test_discard_route_next_hop_down_iface(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[
            IPV4_ROUTE_IFACE_NAME
        ].state = InterfaceState.DOWN

        ipv4_route = gen_ipv4_route()
        ipv6_route = gen_ipv6_route()
        state = RouteState(
            ifaces,
            [],
            {Route.CONFIG: [ipv4_route.to_dict(), ipv6_route.to_dict()]},
        )
        assert [IPV6_ROUTE_IFACE_NAME] == list(
            state.config_iface_routes.keys()
        )
        assert state.config_iface_routes[IPV6_ROUTE_IFACE_NAME] == set(
            [ipv6_route]
        )

    def test_discard_route_next_hop_ipv4_disabled_iface(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME].raw[Interface.IPV4][
            InterfaceIPv4.ENABLED
        ] = False

        ipv4_route = gen_ipv4_route()
        ipv6_route = gen_ipv6_route()
        state = RouteState(
            ifaces,
            [],
            {Route.CONFIG: [ipv4_route.to_dict(), ipv6_route.to_dict()]},
        )
        assert [IPV6_ROUTE_IFACE_NAME] == list(
            state.config_iface_routes.keys()
        )
        assert state.config_iface_routes[IPV6_ROUTE_IFACE_NAME] == set(
            [ipv6_route]
        )

    def test_discard_route_next_hop_ipv6_disabled_iface(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME].raw[Interface.IPV6][
            InterfaceIPv6.ENABLED
        ] = False

        ipv4_route = gen_ipv4_route()
        ipv6_route = gen_ipv6_route()
        state = RouteState(
            ifaces,
            [],
            {Route.CONFIG: [ipv4_route.to_dict(), ipv6_route.to_dict()]},
        )
        assert [IPV4_ROUTE_IFACE_NAME] == list(
            state.config_iface_routes.keys()
        )
        assert state.config_iface_routes[IPV4_ROUTE_IFACE_NAME] == set(
            [ipv4_route]
        )

    def test_discard_route_next_hop_dhcpv4_iface(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME].raw[Interface.IPV4][
            InterfaceIPv4.DHCP
        ] = True

        ipv4_route = gen_ipv4_route()
        ipv6_route = gen_ipv6_route()
        state = RouteState(
            ifaces,
            [],
            {Route.CONFIG: [ipv4_route.to_dict(), ipv6_route.to_dict()]},
        )
        assert [IPV6_ROUTE_IFACE_NAME] == list(
            state.config_iface_routes.keys()
        )
        assert state.config_iface_routes[IPV6_ROUTE_IFACE_NAME] == set(
            [ipv6_route]
        )

    def test_discard_route_next_hop_dhcpv6_iface(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME].raw[Interface.IPV6][
            InterfaceIPv6.DHCP
        ] = True

        ipv4_route = gen_ipv4_route()
        ipv6_route = gen_ipv6_route()
        state = RouteState(
            ifaces,
            [],
            {Route.CONFIG: [ipv4_route.to_dict(), ipv6_route.to_dict()]},
        )
        assert [IPV4_ROUTE_IFACE_NAME] == list(
            state.config_iface_routes.keys()
        )
        assert state.config_iface_routes[IPV4_ROUTE_IFACE_NAME] == set(
            [ipv4_route]
        )

    def test_discard_route_next_hop_ipv6_autoconf_iface(self):
        ifaces = self._gen_ifaces()
        ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME].raw[Interface.IPV6][
            InterfaceIPv6.AUTOCONF
        ] = True

        ipv4_route = gen_ipv4_route()
        ipv6_route = gen_ipv6_route()
        state = RouteState(
            ifaces,
            [],
            {Route.CONFIG: [ipv4_route.to_dict(), ipv6_route.to_dict()]},
        )
        assert [IPV4_ROUTE_IFACE_NAME] == list(
            state.config_iface_routes.keys()
        )
        assert state.config_iface_routes[IPV4_ROUTE_IFACE_NAME] == set(
            [ipv4_route]
        )

    def test_gen_metadata(self):
        ifaces = self._gen_ifaces()
        ipv4_route = gen_ipv4_route()
        ipv6_route = gen_ipv6_route()
        route_state = RouteState(
            ifaces,
            [],
            {Route.CONFIG: [ipv4_route.to_dict(), ipv6_route.to_dict()]},
        )
        ifaces.gen_route_metadata(route_state)

        ipv4_route_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]
        ipv6_route_iface = ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME]

        assert ipv4_route_iface.to_dict()[Interface.IPV4][
            BaseIface.ROUTES_METADATA
        ] == [ipv4_route.to_dict()]
        assert ipv6_route_iface.to_dict()[Interface.IPV6][
            BaseIface.ROUTES_METADATA
        ] == [ipv6_route.to_dict()]


def _create_route(dest, via_addr, via_iface, table, metric):
    return RouteEntry(
        _create_route_dict(dest, via_addr, via_iface, table, metric)
    )


def _create_route_dict(dest, via_addr, via_iface, table, metric):
    return {
        Route.DESTINATION: dest,
        Route.METRIC: metric,
        Route.NEXT_HOP_ADDRESS: via_addr,
        Route.NEXT_HOP_INTERFACE: via_iface,
        Route.TABLE_ID: table,
    }



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import VLAN


@contextmanager
def vlan_interface(ifname, vlan_id, base_iface):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: ifname,
                Interface.TYPE: InterfaceType.VLAN,
                Interface.STATE: InterfaceState.UP,
                VLAN.TYPE: {VLAN.ID: vlan_id, VLAN.BASE_IFACE: base_iface},
            }
        ]
    }
    libnmstate.apply(desired_state)
    try:
        yield desired_state
    finally:
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: ifname,
                        Interface.TYPE: InterfaceType.VLAN,
                        Interface.STATE: InterfaceState.ABSENT,
                    }
                ]
            }
        )



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager

import libnmstate
from libnmstate import schema

from . import statelib


def ifaces_init(*ifnames):
    """ Remove any existing definitions on the interfaces. """
    for ifname in ifnames:
        _set_eth_admin_state(ifname, schema.InterfaceState.ABSENT)


@contextmanager
def iface_up(ifname):
    _set_eth_admin_state(ifname, schema.InterfaceState.UP)
    try:
        yield statelib.show_only((ifname,))
    finally:
        _set_eth_admin_state(ifname, schema.InterfaceState.ABSENT)


def _set_eth_admin_state(ifname, state):
    libnmstate.apply(
        {
            schema.Interface.KEY: [
                {schema.Interface.NAME: ifname, schema.Interface.STATE: state}
            ]
        }
    )


def get_mac_address(ifname):
    state = statelib.show_only((ifname,))
    return state[schema.Interface.KEY][0].get(schema.Interface.MAC)



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
""" Test the nmstate example files """
import os

import pytest

from .testlib import assertlib
from .testlib.examplelib import example_state

from libnmstate import netinfo
from libnmstate.error import NmstateNotSupportedError
from libnmstate.schema import DNS


@pytest.mark.tier1
def test_add_down_remove_vlan(eth1_up):
    """
    Test adding, downing and removing a vlan
    """

    vlan_ifname = "eth1.101"
    with example_state(
        "vlan101_eth1_up.yml", cleanup="vlan101_eth1_absent.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)
        with example_state("vlan101_eth1_down.yml") as desired_state:
            assertlib.assert_absent(vlan_ifname)

    assertlib.assert_absent(vlan_ifname)


@pytest.mark.tier1
def test_add_remove_ovs_bridge(eth1_up):
    with example_state(
        "ovsbridge_create.yml", cleanup="ovsbridge_delete.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("ovs-br0")


@pytest.mark.tier1
def test_add_remove_ovs_bridge_bond(eth1_up, eth2_up):
    with example_state(
        "ovsbridge_bond_create.yml", cleanup="ovsbridge_delete.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("ovs-br0")
    assertlib.assert_absent("ovs-bond1")


@pytest.mark.tier1
def test_add_remove_ovs_bridge_vlan(eth1_up, eth2_up):
    with example_state(
        "ovsbridge_vlan_port.yml", cleanup="ovsbridge_delete.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("ovs-br0")


@pytest.mark.tier1
def test_add_remove_linux_bridge(eth1_up):
    with example_state(
        "linuxbrige_eth1_up.yml", cleanup="linuxbrige_eth1_absent.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("linux-br0")


@pytest.mark.tier1
def test_bond_linuxbridge_vlan(eth1_up, eth2_up):
    with example_state(
        "bond_linuxbridge_vlan_up.yml",
        cleanup="bond_linuxbridge_vlan_absent.yml",
    ) as desired_state:
        assertlib.assert_state_match(desired_state)

    assertlib.assert_absent("bond0")
    assertlib.assert_absent("br0")
    assertlib.assert_absent("br29")
    assertlib.assert_absent("vlan29")


@pytest.mark.tier1
def test_dns_edit(eth1_up):
    with example_state(
        "dns_edit_eth1.yml", cleanup="dns_remove.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    current_state = netinfo.show()
    assert current_state.get(DNS.KEY, {}).get(DNS.CONFIG, {}) == {}


@pytest.mark.tier1
def test_add_remove_routes(eth1_up):
    """
    Test adding a strict route and removing all routes next hop to eth1.
    """
    with example_state(
        "eth1_add_route.yml", cleanup="eth1_del_all_routes.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_no_config_route_to_iface("eth1")


@pytest.mark.tier1
@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="Team kmod not available in Travis CI",
)
def test_add_remove_team_with_port(eth1_up, eth2_up):
    with example_state(
        "team0_with_port.yml", cleanup="team0_absent.yml"
    ) as desired_state:
        assertlib.assert_state_match(desired_state)

    assertlib.assert_absent("team0")


@pytest.mark.skipif(
    os.environ.get("CI") == "true",
    reason="SR-IOV device required for this test case",
)
@pytest.mark.xfail(
    raises=NmstateNotSupportedError,
    reason="The device does not support SR-IOV.",
)
def test_set_ethernet_sriov(eth1_up):
    with example_state("eth1_with_sriov.yml") as desired_state:
        assertlib.assert_state_match(desired_state)


def test_port_vlan(eth1_up):
    with example_state(
        "linuxbrige_eth1_up_port_vlan.yml",
        cleanup="linuxbrige_eth1_absent.yml",
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("linux-br0")


def test_add_ovs_patch_and_remove():
    with example_state(
        "ovsbridge_patch_create.yml", cleanup="ovsbridge_patch_delete.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("patch0")


def test_add_remove_vrf(eth1_up):
    with example_state(
        "vrf0_with_eth1.yml", cleanup="vrf0_absent.yml"
    ) as desired_state:
        assertlib.assert_state_match(desired_state)

    assertlib.assert_absent("vrf0")


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for infiniband test",
)
def test_add_ib_pkey_nic_and_remove():
    test_nic = os.environ["TEST_REAL_NIC"]
    with example_state(
        "infiniband_pkey_ipoib_create.yml",
        cleanup="infiniband_pkey_ipoib_delete.yml",
        substitute=("mlx5_ib0", test_nic),
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent(f"{test_nic}.80ff")


def test_add_mac_vlan_and_remove():
    with example_state(
        "mac_vlan_create.yml", cleanup="mac_vlan_absent.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("macvlan0")


def test_add_mac_vtap_and_remove():
    with example_state(
        "mac_vtap_create.yml", cleanup="mac_vtap_absent.yml"
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent("macvtap0")



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

import libnmstate

from libnmstate.error import NmstateNotSupportedError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import VRF

from .testlib import assertlib
from .testlib import statelib

TEST_VRF0 = "test-vrf0"
TEST_VRF1 = "test-vrf1"
TEST_VRF_PORT0 = "eth1"
TEST_VRF_PORT1 = "eth2"
TEST_ROUTE_TABLE_ID0 = 100
TEST_ROUTE_TABLE_ID1 = 101
IPV4_ADDRESS1 = "192.0.2.251"
IPV6_ADDRESS1 = "2001:db8:1::1"


@pytest.fixture
def vrf0_with_port0(port1_up):
    vrf_iface_info = {
        Interface.NAME: TEST_VRF0,
        Interface.TYPE: InterfaceType.VRF,
        VRF.CONFIG_SUBTREE: {
            VRF.PORT_SUBTREE: [TEST_VRF_PORT0],
            VRF.ROUTE_TABLE_ID: TEST_ROUTE_TABLE_ID0,
        },
    }
    libnmstate.apply({Interface.KEY: [vrf_iface_info]})
    yield vrf_iface_info
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_VRF0,
                    Interface.STATE: InterfaceState.ABSENT,
                }
            ]
        }
    )
    assertlib.assert_absent(TEST_VRF0)


@pytest.fixture
def vrf1_with_port1(port1_up):
    vrf_iface_info = {
        Interface.NAME: TEST_VRF1,
        Interface.TYPE: InterfaceType.VRF,
        VRF.CONFIG_SUBTREE: {
            VRF.PORT_SUBTREE: [TEST_VRF_PORT1],
            VRF.ROUTE_TABLE_ID: TEST_ROUTE_TABLE_ID1,
        },
    }
    libnmstate.apply({Interface.KEY: [vrf_iface_info]})
    yield vrf_iface_info
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_VRF1,
                    Interface.STATE: InterfaceState.ABSENT,
                }
            ]
        }
    )
    assertlib.assert_absent(TEST_VRF1)


class TestVrf:
    def test_create_and_remove(self, vrf0_with_port0):
        pass

    def test_change_route_table_id(self, vrf0_with_port0):
        iface_info = vrf0_with_port0
        iface_info[VRF.CONFIG_SUBTREE][VRF.ROUTE_TABLE_ID] += 1
        desired_state = {Interface.KEY: [iface_info]}
        with pytest.raises(NmstateNotSupportedError):
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)

    def test_create_with_empty_ports(self):
        desired_state = {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_VRF0,
                    Interface.TYPE: InterfaceType.VRF,
                    VRF.CONFIG_SUBTREE: {
                        VRF.PORT_SUBTREE: [],
                        VRF.ROUTE_TABLE_ID: TEST_ROUTE_TABLE_ID0,
                    },
                }
            ]
        }
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_add_and_remove_port(self, vrf0_with_port0, port1_up):
        iface_info = vrf0_with_port0
        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE].append(TEST_VRF_PORT1)
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE].remove(TEST_VRF_PORT1)
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_remove_port(self, vrf0_with_port0):
        iface_info = vrf0_with_port0
        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = []
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_remove_all_ports(self, vrf0_with_port0):
        iface_info = vrf0_with_port0
        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = []
        desired_state = {Interface.KEY: [iface_info]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_moving_port_from_other_vrf(
        self, vrf0_with_port0, vrf1_with_port1
    ):
        vrf0_iface = vrf0_with_port0
        vrf1_iface = vrf1_with_port1
        vrf0_iface[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = []
        vrf1_iface[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = [
            TEST_VRF_PORT0,
            TEST_VRF_PORT1,
        ]

        desired_state = {Interface.KEY: [vrf0_iface, vrf1_iface]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_swaping_port(self, vrf0_with_port0, vrf1_with_port1):
        vrf0_iface = vrf0_with_port0
        vrf1_iface = vrf1_with_port1
        vrf0_iface[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = [TEST_VRF_PORT1]
        vrf1_iface[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = [TEST_VRF_PORT0]
        desired_state = {Interface.KEY: [vrf0_iface, vrf1_iface]}
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_port_holding_ip(self, vrf0_with_port0):
        desired_state = {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_VRF_PORT0,
                    Interface.IPV4: {
                        InterfaceIPv4.ENABLED: True,
                        InterfaceIPv4.DHCP: False,
                        InterfaceIPv4.ADDRESS: [
                            {
                                InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                            }
                        ],
                    },
                    Interface.IPV6: {
                        InterfaceIPv6.ENABLED: True,
                        InterfaceIPv6.DHCP: False,
                        InterfaceIPv6.AUTOCONF: False,
                        InterfaceIPv6.ADDRESS: [
                            {
                                InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                            }
                        ],
                    },
                }
            ]
        }
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)

    def test_show_saved_config_with_vrf_down(self, vrf0_with_port0):
        running_state = statelib.show_only((TEST_VRF0,))
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: TEST_VRF0,
                        Interface.STATE: InterfaceState.DOWN,
                    }
                ]
            }
        )
        saved_state = statelib.show_saved_config_only((TEST_VRF0,))
        iface_state = saved_state[Interface.KEY][0]

        assert (
            saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
        )
        assert (
            iface_state[VRF.CONFIG_SUBTREE][VRF.ROUTE_TABLE_ID]
            == TEST_ROUTE_TABLE_ID0
        )
        assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
""" Use a Network Manager User Setting to store and retrieve information that
does not fit somewhere else such as an interface's description.

https://lazka.github.io/pgi-docs/#NM-1.0/classes/SettingUser.html
"""

from libnmstate.error import NmstateValueError
from .common import NM

NMSTATE_DESCRIPTION = "nmstate.interface.description"


def create_setting(iface_state, base_con_profile):
    description = iface_state.get("description")

    if not description:
        return None

    if not NM.SettingUser.check_val(description):
        raise NmstateValueError("Invalid description")

    user_setting = None
    if base_con_profile:
        user_setting = base_con_profile.get_setting_by_name(
            NM.SETTING_USER_SETTING_NAME
        )
        if user_setting:
            user_setting = user_setting.duplicate()

    if not user_setting:
        user_setting = NM.SettingUser.new()

    user_setting.set_data(NMSTATE_DESCRIPTION, description)
    return user_setting


def get_user_config(nm_profile, full_config=False):
    """
    Get description from user settings for a connection
    """
    info = {}
    if nm_profile:
        user_setting = nm_profile.get_setting_by_name(
            NM.SETTING_USER_SETTING_NAME
        )
        if user_setting:
            description = user_setting.get_data(NMSTATE_DESCRIPTION)
            if description:
                info["description"] = description

    return info



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.prettystate import PrettyState

TEST_STATE = {
    "0": "a",
    "type": "unknown",
    "name": "foo",
    "ipv4": {"abc": 1, "enabled": "false"},
    "port": [{"a": 1, "name": "bar1"}, {"a": 2, "name": "bar2"}],
    "state": "up",
}

TEST_YAML_STATE = """---
name: foo
type: unknown
state: up
'0': a
ipv4:
  enabled: 'false'
  abc: 1
port:
- name: bar1
  a: 1
- name: bar2
  a: 2
"""

TEST_JSON_STATE = """{
    "name": "foo",
    "type": "unknown",
    "state": "up",
    "0": "a",
    "ipv4": {
        "enabled": "false",
        "abc": 1
    },
    "port": [
        {
            "name": "bar1",
            "a": 1
        },
        {
            "name": "bar2",
            "a": 2
        }
    ]
}"""


def test_pretty_state_order_with_priority():
    state = PrettyState(TEST_STATE)
    assert state.yaml == TEST_YAML_STATE


def test_json_ovs_bond_name_list_first():
    state = PrettyState(TEST_STATE)
    assert state.json == TEST_JSON_STATE



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

from libnmstate.error import NmstateValueError
from libnmstate.error import NmstateVerificationError
from libnmstate.error import NmstateNotImplementedError
from libnmstate.iplib import is_ipv6_address
from libnmstate.prettystate import format_desired_current_state_diff
from libnmstate.schema import DNS
from libnmstate.schema import Interface


class DnsState:
    PRIORITY_METADATA = "_priority"

    def __init__(self, des_dns_state, cur_dns_state):
        self._config_changed = False
        if des_dns_state is None or des_dns_state.get(DNS.CONFIG) is None:
            # Use current config if DNS.KEY not defined or DNS.CONFIG not
            # defined.
            self._dns_state = cur_dns_state or {}
        else:
            self._dns_state = des_dns_state
            self._validate()
            self._config_changed = _is_dns_config_changed(
                des_dns_state, cur_dns_state
            )
        self._cur_dns_state = deepcopy(cur_dns_state) if cur_dns_state else {}

    @property
    def current_config(self):
        return _get_config(self._cur_dns_state)

    @property
    def config(self):
        return _get_config(self._dns_state)

    @property
    def _config_servers(self):
        return _get_config_servers(self._dns_state)

    @property
    def _config_searches(self):
        return _get_config_searches(self._dns_state)

    def gen_metadata(self, ifaces, route_state):
        """
        Return DNS configure targeting to store as metadata of interface.
        Data structure returned is:
            {
                iface_name: {
                    Interface.IPV4: {
                        DNS.SERVER: dns_servers,
                        DNS.SEARCH: dns_searches,
                    },
                    Interface.IPV6: {
                        DNS.SERVER: dns_servers,
                        DNS.SEARCH: dns_searches,
                    },
                }
            }
        """
        iface_metadata = {}
        if not self._config_servers and not self._config_searches:
            return iface_metadata
        ipv4_iface, ipv6_iface = self._find_ifaces_for_name_servers(
            ifaces, route_state
        )
        if ipv4_iface == ipv6_iface:
            iface_metadata = {
                ipv4_iface: {
                    Interface.IPV4: {DNS.SERVER: [], DNS.SEARCH: []},
                    Interface.IPV6: {DNS.SERVER: [], DNS.SEARCH: []},
                },
            }
        else:
            if ipv4_iface:
                iface_metadata[ipv4_iface] = {
                    Interface.IPV4: {DNS.SERVER: [], DNS.SEARCH: []},
                }
            if ipv6_iface:
                iface_metadata[ipv6_iface] = {
                    Interface.IPV6: {DNS.SERVER: [], DNS.SEARCH: []},
                }
        index = 0
        searches_saved = False
        for server in self._config_servers:
            iface_name = None
            if is_ipv6_address(server):
                iface_name = ipv6_iface
                family = Interface.IPV6
            else:
                iface_name = ipv4_iface
                family = Interface.IPV4
            if not iface_name:
                raise NmstateValueError(
                    "Failed to find suitable interface for saving DNS "
                    "name servers: %s" % server
                )
            iface_dns_metada = iface_metadata[iface_name][family]
            iface_dns_metada[DNS.SERVER].append(server)
            iface_dns_metada.setdefault(DnsState.PRIORITY_METADATA, index)
            if not searches_saved:
                iface_dns_metada[DNS.SEARCH] = self._config_searches
            searches_saved = True
            index += 1
        return iface_metadata

    def _find_ifaces_for_name_servers(self, ifaces, route_state):
        """
        Find interface to store the DNS configurations in the order of:
            * Any interface with static gateway
            * Any interface configured as dynamic IP with 'auto-dns:False'
        Return tuple: (ipv4_iface, ipv6_iface)
        """
        ipv4_iface, ipv6_iface = self._find_ifaces_with_static_gateways(
            route_state
        )
        if not (ipv4_iface and ipv6_iface):
            (
                auto_ipv4_iface,
                auto_ipv6_iface,
            ) = self._find_ifaces_with_auto_dns_false(ifaces)
            if not ipv4_iface and auto_ipv4_iface:
                ipv4_iface = auto_ipv4_iface
            if not ipv6_iface and auto_ipv6_iface:
                ipv6_iface = auto_ipv6_iface

        return ipv4_iface, ipv6_iface

    def _find_ifaces_with_static_gateways(self, route_state):
        """
        Return tuple of interfaces with IPv4 and IPv6 static gateways.
        """
        ipv4_iface = None
        ipv6_iface = None
        for iface_name, route_set in route_state.config_iface_routes.items():
            for route in route_set:
                if ipv4_iface and ipv6_iface:
                    return (ipv4_iface, ipv6_iface)
                if route.is_gateway:
                    if route.is_ipv6:
                        ipv6_iface = iface_name
                    else:
                        ipv4_iface = iface_name
        return (ipv4_iface, ipv6_iface)

    def _find_ifaces_with_auto_dns_false(self, ifaces):
        ipv4_iface = None
        ipv6_iface = None
        for iface in ifaces.all_kernel_ifaces.values():
            if ipv4_iface and ipv6_iface:
                return (ipv4_iface, ipv6_iface)
            for family in (Interface.IPV4, Interface.IPV6):
                ip_state = iface.ip_state(family)
                if ip_state.is_dynamic and (not ip_state.auto_dns):
                    if family == Interface.IPV4:
                        ipv4_iface = iface.name
                    else:
                        ipv6_iface = iface.name

        return (ipv4_iface, ipv6_iface)

    def verify(self, cur_dns_state):
        cur_dns = DnsState(
            des_dns_state=None,
            cur_dns_state=cur_dns_state,
        )
        if self.config.get(DNS.SERVER, []) != cur_dns.config.get(
            DNS.SERVER, []
        ) or self.config.get(DNS.SEARCH, []) != cur_dns.config.get(
            DNS.SEARCH, []
        ):
            raise NmstateVerificationError(
                format_desired_current_state_diff(
                    {DNS.KEY: self.config},
                    {DNS.KEY: cur_dns.config},
                )
            )

    def _validate(self):
        if (
            len(self._config_servers) > 2
            and any(is_ipv6_address(n) for n in self._config_servers)
            and any(not is_ipv6_address(n) for n in self._config_servers)
            and _is_mixed_dns_servers(self._config_servers)
        ):
            raise NmstateNotImplementedError(
                "Placing IPv4/IPv6 nameserver in the middle of IPv6/IPv4 "
                "nameservers is not supported yet"
            )

    @property
    def config_changed(self):
        return self._config_changed


def _get_config(state):
    conf = state.get(DNS.CONFIG, {})
    if not conf:
        conf = {DNS.SERVER: [], DNS.SEARCH: []}
    return conf


def _get_config_servers(state):
    return _get_config(state).get(DNS.SERVER, [])


def _get_config_searches(state):
    return _get_config(state).get(DNS.SEARCH, [])


def _is_dns_config_changed(des_dns_state, cur_dns_state):
    return _get_config_servers(des_dns_state) != _get_config_servers(
        cur_dns_state
    ) or _get_config_searches(des_dns_state) != _get_config_searches(
        cur_dns_state
    )


def _is_mixed_dns_servers(servers):
    """
    Return True when an IPv6 server is in the middle of two IPv4 namesevers or
    an IPv4 server is in the middle of two IPv6 servers.
    """
    pattern = ""
    for server in servers:
        if is_ipv6_address(server):
            pattern += "6"
        else:
            pattern += "4"

    return "464" in pattern or "646" in pattern



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#


class NmstateError(Exception):
    """
    The base exception of libnmstate.
    """

    pass


class NmstateDependencyError(NmstateError):
    """
    Nmstate requires external tools installed and/or started for desired state.
    """

    pass


class NmstateValueError(NmstateError, ValueError):
    """
    Exception happens at pre-apply check, user should resubmit the amended
    desired state. Example:
        * JSON/YAML syntax issue.
        * Nmstate schema issue.
        * Invalid value of desired property, like bond missing port.
    """

    pass


class NmstatePermissionError(NmstateError, PermissionError):
    """
    Permission deny when applying the desired state.
    """

    pass


class NmstateConflictError(NmstateError, RuntimeError):
    """
    Something else is already editing the network state via Nmstate.
    """

    pass


class NmstateLibnmError(NmstateError):
    """
    Exception for unexpected libnm failure.
    """

    pass


class NmstateVerificationError(NmstateError):
    """
    After applied desired state, current state does not match desired state for
    unknown reason.
    """

    pass


class NmstateKernelIntegerRoundedError(NmstateVerificationError):
    """
    After applied desired state, current state does not match desire state
    due to integer been rounded by kernel.
    For example, with HZ configured as 250 in kernel, the linux bridge option
    multicast_startup_query_interval, 3125 will be rounded to 3124.
    """

    pass


class NmstateNotImplementedError(NmstateError, NotImplementedError):
    """
    Desired feature is not supported by Nmstate yet.
    """

    pass


class NmstateInternalError(NmstateError):
    """
    Unexpected behaviour happened. It is a bug of libnmstate which should be
    fixed.
    """

    pass


class NmstateNotSupportedError(NmstateError):
    """
    A resource like a device does not support the requested feature.
    """

    pass


class NmstateTimeoutError(NmstateLibnmError):
    """
    The transaction execution timed out.
    """

    pass


class NmstatePluginError(NmstateError):
    """
    Unexpected plugin behaviour happens, it is a bug of the plugin.
    """

    pass



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager

import pytest

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import MacVtap

from .testlib import assertlib
from .testlib import statelib


ETH1 = "eth1"
MACVLAN0 = "macvtap0"
MACVTAP0 = "macvtap0"


@pytest.mark.parametrize(
    "mode",
    [
        MacVtap.Mode.VEPA,
        MacVtap.Mode.BRIDGE,
        MacVtap.Mode.PRIVATE,
        MacVtap.Mode.PASSTHRU,
    ],
)
def test_add_mac_vtap_multiple_modes(eth1_up, mode):
    with macvtap_interface(MACVLAN0, mode, True) as desired_state:
        assertlib.assert_state(desired_state)
    assertlib.assert_absent(MACVLAN0)


def test_add_mac_vtap_promiscuous_off(eth1_up):
    with macvtap_interface(
        MACVLAN0, MacVtap.Mode.PASSTHRU, False
    ) as desired_state:
        libnmstate.apply(desired_state)
    assertlib.assert_absent(MACVLAN0)


def test_edit_mac_vtap(eth1_up):
    with macvtap_interface(
        MACVLAN0, MacVtap.Mode.PASSTHRU, True
    ) as desired_state:
        assertlib.assert_state(desired_state)
        desired_state[Interface.KEY][0][Interface.MTU] = 1400
        libnmstate.apply(desired_state)
        assertlib.assert_state(desired_state)

    assertlib.assert_absent(MACVLAN0)


@contextmanager
def macvtap_interface(ifname, mode, promiscuous):
    d_state = {
        Interface.KEY: [
            {
                Interface.NAME: ifname,
                Interface.TYPE: MacVtap.TYPE,
                Interface.STATE: InterfaceState.UP,
                MacVtap.CONFIG_SUBTREE: {
                    MacVtap.BASE_IFACE: ETH1,
                    MacVtap.MODE: mode,
                    MacVtap.PROMISCUOUS: promiscuous,
                },
            }
        ]
    }
    try:
        libnmstate.apply(d_state)
        yield d_state
    finally:
        d_state[Interface.KEY][0][Interface.STATE] = InterfaceState.ABSENT
        libnmstate.apply(d_state)


@pytest.fixture
def macvtap0_up(eth1_up):
    with macvtap_interface(MACVTAP0, MacVtap.Mode.PASSTHRU, True):
        yield


def test_show_saved_config_with_macvtap_down(macvtap0_up):
    running_state = statelib.show_only((MACVTAP0,))
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: MACVTAP0,
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only((MACVTAP0,))

    assert saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
    assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import pytest

import libnmstate
from libnmstate.prettystate import PrettyState
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import OVSBridge
from libnmstate.schema import OvsDB
from libnmstate.schema import OVSInterface
from libnmstate.state import state_match
from libnmstate.error import NmstateDependencyError
from libnmstate.error import NmstateNotSupportedError
from libnmstate.error import NmstateValueError

from .testlib import assertlib
from .testlib import cmdlib
from .testlib import statelib
from .testlib.env import nm_major_minor_version
from .testlib.nmplugin import disable_nm_plugin
from .testlib.ovslib import Bridge
from .testlib.servicelib import disable_service
from .testlib.ovslib import get_proxy_port_profile_name_of_ovs_interface
from .testlib.ovslib import get_nm_active_profiles
from .testlib.vlan import vlan_interface


BOND1 = "bond1"
BRIDGE0 = "br0"
BRIDGE1 = "br1"
PORT1 = "ovs1"
PORT2 = "ovs2"
PATCH0 = "patch0"
PATCH1 = "patch1"
VLAN_IFNAME = "eth101"

MAC1 = "02:FF:FF:FF:FF:01"

ETH1 = "eth1"
ETH2 = "eth2"

OVS_BOND_YAML_STATE = f"""
    port:
    - name: {BOND1}
      link-aggregation:
        mode: active-backup
        port:
        - name: {ETH1}
        - name: {ETH2}
"""

RC_SUCCESS = 0
TEST_EXTERNAL_IDS_KEY = "ovn-bridge-mappings"
TEST_EXTERNAL_IDS_VALUE = "provider:br-provider"
TEST_OTHER_CONFIG_KEY = "stats-update-interval"
TEST_OTHER_CONFIG_VALUE = "1000"


@pytest.fixture
def bridge_with_ports(port0_up):
    system_port0_name = port0_up[Interface.KEY][0][Interface.NAME]

    bridge = Bridge(BRIDGE1)
    bridge.add_system_port(system_port0_name)
    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})
    with bridge.create():
        yield bridge


@pytest.mark.tier1
def test_create_and_remove_ovs_bridge_with_min_desired_state():
    with Bridge(BRIDGE1).create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)


def test_create_and_remove_ovs_bridge_options_specified():
    bridge = Bridge(BRIDGE1)
    bridge.set_options(
        {
            OVSBridge.Options.FAIL_MODE: "",
            OVSBridge.Options.MCAST_SNOOPING_ENABLED: False,
            OVSBridge.Options.RSTP: False,
            OVSBridge.Options.STP: True,
        }
    )

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)


@pytest.mark.tier1
def test_create_and_remove_ovs_bridge_with_a_system_port(port0_up):
    bridge = Bridge(BRIDGE1)
    port0_name = port0_up[Interface.KEY][0][Interface.NAME]
    bridge.add_system_port(port0_name)

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)

    state = statelib.show_only((port0_name,))
    assert state
    assert state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP


@pytest.mark.tier1
def test_create_and_remove_ovs_bridge_with_internal_port_static_ip_and_mac():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(
        PORT1,
        mac=MAC1,
        ipv4_state={
            InterfaceIPv4.ENABLED: True,
            InterfaceIPv4.ADDRESS: [
                {
                    InterfaceIPv4.ADDRESS_IP: "192.0.2.1",
                    InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                }
            ],
        },
    )

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


@pytest.fixture
def bridge_with_same_name_iface():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(
        BRIDGE1, ipv4_state={InterfaceIPv4.ENABLED: False}
    )

    with bridge.create() as state:
        yield state


@pytest.mark.tier1
def test_create_and_remove_ovs_bridge_with_internal_port_same_name(
    bridge_with_same_name_iface,
):
    state = statelib.show_only((BRIDGE1,))
    assert state
    assert len(state[Interface.KEY]) == 2


@pytest.mark.tier1
def test_vlan_as_ovs_bridge_port(vlan_on_eth1):
    bridge = Bridge(BRIDGE1)
    bridge.add_system_port(vlan_on_eth1)
    with bridge.create() as state:
        assertlib.assert_state_match(state)


@pytest.mark.tier1
def test_ovs_interface_with_max_length_name():
    bridge = Bridge(BRIDGE1)
    ovs_interface_name = "ovs123456789012"
    bridge.add_internal_port(ovs_interface_name)

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(ovs_interface_name)


def test_nm_ovs_plugin_missing():
    with disable_nm_plugin("ovs"):
        with pytest.raises(NmstateDependencyError):
            libnmstate.apply(
                {
                    Interface.KEY: [
                        {
                            Interface.NAME: BRIDGE1,
                            Interface.TYPE: InterfaceType.OVS_BRIDGE,
                            Interface.STATE: InterfaceState.UP,
                        }
                    ]
                }
            )


def test_ovs_service_missing():
    with disable_service("openvswitch"):
        with pytest.raises(NmstateDependencyError):
            libnmstate.apply(
                {
                    Interface.KEY: [
                        {
                            Interface.NAME: BRIDGE1,
                            Interface.TYPE: InterfaceType.OVS_BRIDGE,
                            Interface.STATE: InterfaceState.UP,
                        }
                    ]
                }
            )

    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: BRIDGE1,
                    Interface.STATE: InterfaceState.ABSENT,
                }
            ]
        }
    )


class _OvsProfileStillExists(Exception):
    pass


@pytest.mark.tier1
def test_remove_ovs_internal_iface_got_port_profile_removed(bridge_with_ports):
    for port_name in bridge_with_ports.ports_names:
        active_profiles = get_nm_active_profiles()
        assert port_name in active_profiles
        proxy_port_profile = get_proxy_port_profile_name_of_ovs_interface(
            port_name
        )
        assert proxy_port_profile
        assert proxy_port_profile in active_profiles
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: port_name,
                        Interface.STATE: InterfaceState.ABSENT,
                    }
                ]
            }
        )

        rc, output, _ = cmdlib.exec_cmd(
            f"nmcli connection show {proxy_port_profile}".split(" "),
        )
        if rc == 0:
            raise _OvsProfileStillExists(
                f"{proxy_port_profile} still exists: {output}"
            )


@pytest.fixture
def vlan_on_eth1(eth1_up):
    with vlan_interface(
        VLAN_IFNAME, 101, eth1_up[Interface.KEY][0][Interface.NAME]
    ):
        yield VLAN_IFNAME


@pytest.mark.tier1
def test_change_ovs_interface_mac():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})

    with bridge.create() as state:
        desired_state = {
            Interface.KEY: [{Interface.NAME: PORT1, Interface.MAC: MAC1}]
        }
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


class TestOvsLinkAggregation:
    @pytest.mark.tier1
    def test_create_and_remove_lag(self, port0_up, port1_up):
        port0_name = port0_up[Interface.KEY][0][Interface.NAME]
        port1_name = port1_up[Interface.KEY][0][Interface.NAME]

        bridge = Bridge(BRIDGE1)
        bridge.add_link_aggregation_port(BOND1, (port0_name, port1_name))

        with bridge.create() as state:
            assertlib.assert_state_match(state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BOND1)

    @pytest.mark.tier1
    def test_create_lag_with_ports_in_reverse_order(self, port0_up, port1_up):
        port0_name = port0_up[Interface.KEY][0][Interface.NAME]
        port1_name = port1_up[Interface.KEY][0][Interface.NAME]

        bridge = Bridge(BRIDGE1)
        bridge.add_link_aggregation_port(BOND1, (port1_name, port0_name))

        with bridge.create() as state:
            assertlib.assert_state_match(state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BOND1)

    def test_pretty_state_ovs_lag_name_first(self, eth1_up, eth2_up):
        bridge = Bridge(BRIDGE1)
        bridge.add_link_aggregation_port(BOND1, (ETH1, ETH2))

        with bridge.create():
            current_state = statelib.show_only((BRIDGE1,))
            pretty_state = PrettyState(current_state)
            assert OVS_BOND_YAML_STATE in pretty_state.yaml

    @pytest.mark.tier1
    def test_add_ovs_lag_to_existing_ovs_bridge(self, port0_up, port1_up):
        with Bridge(BRIDGE1).create():
            port0_name = port0_up[Interface.KEY][0][Interface.NAME]
            port1_name = port1_up[Interface.KEY][0][Interface.NAME]
            bridge = Bridge(BRIDGE1)
            bridge.add_link_aggregation_port(BOND1, (port1_name, port0_name))
            libnmstate.apply(bridge.state)
            assertlib.assert_state_match(bridge.state)


@pytest.mark.tier1
def test_ovs_vlan_access_tag():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT1, ipv4_state={InterfaceIPv4.ENABLED: False})
    bridge.set_port_option(
        PORT1,
        {
            OVSBridge.Port.NAME: PORT1,
            OVSBridge.Port.VLAN_SUBTREE: {
                OVSBridge.Port.Vlan.MODE: OVSBridge.Port.Vlan.Mode.ACCESS,
                OVSBridge.Port.Vlan.TAG: 2,
            },
        },
    )
    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


def test_add_invalid_port_ip_config(eth1_up):
    desired_state = eth1_up
    desired_state[Interface.KEY][0][Interface.IPV4][InterfaceIP.ENABLED] = True
    desired_state[Interface.KEY][0][Interface.IPV4][InterfaceIP.DHCP] = True
    with pytest.raises(NmstateValueError):
        bridge = Bridge(name=BRIDGE1)
        bridge.add_system_port("eth1")
        with bridge.create() as state:
            desired_state[Interface.KEY].append(state[Interface.KEY][0])
            libnmstate.apply(desired_state)


def test_ovsdb_new_bridge_with_external_id():
    bridge = Bridge(BRIDGE1)
    bridge.set_ovs_db({OvsDB.EXTERNAL_IDS: {"foo": "abc", "bak": 1}})
    bridge.add_internal_port(
        PORT1,
        ipv4_state={InterfaceIPv4.ENABLED: False},
        ovs_db={OvsDB.EXTERNAL_IDS: {"foo": "abcd", "bak": 2}},
    )
    with bridge.create() as state:
        assertlib.assert_state_match(state)
        new_state = statelib.show_only((PORT1,))
        # The newly created OVS internal interface should also hold
        # NM created external IDS.
        assert (
            "NM.connection.uuid"
            in new_state[Interface.KEY][0][OvsDB.OVS_DB_SUBTREE][
                OvsDB.EXTERNAL_IDS
            ]
        )


def test_ovsdb_set_external_ids_for_ovs_system_interface(bridge_with_ports):
    system_port_name = "eth1"
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: system_port_name,
                OvsDB.OVS_DB_SUBTREE: {OvsDB.EXTERNAL_IDS: {"foo": 1000}},
            }
        ]
    }
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


def test_ovsdb_set_external_ids_for_existing_bridge(bridge_with_ports):
    bridge = bridge_with_ports
    bridge.set_ovs_db({OvsDB.EXTERNAL_IDS: {"foo": "abc", "bak": 1}})
    bridge.add_internal_port(
        PORT2,
        ipv4_state={InterfaceIPv4.ENABLED: False},
        ovs_db={OvsDB.EXTERNAL_IDS: {"foo": "abcd", "bak": 2}},
    )
    bridge.apply()
    assertlib.assert_state_match(bridge.state)


@pytest.fixture
def ovs_bridge_with_custom_external_ids():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(
        PORT1,
        ipv4_state={InterfaceIPv4.ENABLED: False},
        ovs_db={OvsDB.EXTERNAL_IDS: {"foo": "abcd", "bak": 2}},
    )
    with bridge.create() as state:
        yield state


def test_ovsdb_remove_external_ids(ovs_bridge_with_custom_external_ids):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: PORT1,
                    OvsDB.OVS_DB_SUBTREE: {OvsDB.EXTERNAL_IDS: {}},
                }
            ]
        }
    )
    iface_info = statelib.show_only((PORT1,))[Interface.KEY][0]
    external_ids = iface_info[OvsDB.OVS_DB_SUBTREE][OvsDB.EXTERNAL_IDS]
    assert len(external_ids) == 1


def test_ovsdb_override_external_ids(ovs_bridge_with_custom_external_ids):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: PORT1,
                    OvsDB.OVS_DB_SUBTREE: {
                        OvsDB.EXTERNAL_IDS: {"new_ids": "haha"}
                    },
                }
            ]
        }
    )
    iface_info = statelib.show_only((PORT1,))[Interface.KEY][0]
    external_ids = iface_info[OvsDB.OVS_DB_SUBTREE][OvsDB.EXTERNAL_IDS]
    assert len(external_ids) == 2
    assert external_ids["new_ids"] == "haha"


def test_ovsdb_preserved_if_not_mentioned(ovs_bridge_with_custom_external_ids):
    libnmstate.apply(
        {Interface.KEY: [{Interface.NAME: PORT1, Interface.MTU: 1501}]}
    )
    iface_info = statelib.show_only((PORT1,))[Interface.KEY][0]
    external_ids = iface_info[OvsDB.OVS_DB_SUBTREE][OvsDB.EXTERNAL_IDS]
    assert len(external_ids) > 1


def test_ovsdb_global_config_add_external_ids_and_remove():
    desired_ovs_config = {
        OvsDB.EXTERNAL_IDS: {TEST_EXTERNAL_IDS_KEY: TEST_EXTERNAL_IDS_VALUE}
    }
    libnmstate.apply({OvsDB.KEY: desired_ovs_config})
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert state_match(desired_ovs_config, current_ovs_config)

    libnmstate.apply(
        {OvsDB.KEY: {OvsDB.EXTERNAL_IDS: {TEST_EXTERNAL_IDS_KEY: None}}}
    )
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert TEST_EXTERNAL_IDS_KEY not in current_ovs_config[OvsDB.EXTERNAL_IDS]


def test_ovsdb_global_config_add_other_config_and_remove():
    desired_ovs_config = {
        OvsDB.OTHER_CONFIG: {TEST_OTHER_CONFIG_KEY: TEST_OTHER_CONFIG_VALUE}
    }
    libnmstate.apply({OvsDB.KEY: desired_ovs_config})
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert state_match(desired_ovs_config, current_ovs_config)

    libnmstate.apply(
        {OvsDB.KEY: {OvsDB.OTHER_CONFIG: {TEST_OTHER_CONFIG_KEY: None}}}
    )
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert TEST_OTHER_CONFIG_KEY not in current_ovs_config[OvsDB.OTHER_CONFIG]


def test_remove_all_ovsdb_global_config():
    libnmstate.apply({OvsDB.KEY: {}})
    current_ovs_config = libnmstate.show()[OvsDB.KEY]

    assert current_ovs_config == {
        OvsDB.EXTERNAL_IDS: {},
        OvsDB.OTHER_CONFIG: {},
    }


@pytest.fixture
def ovsdb_global_config_external_ids():
    ovs_config = {
        OvsDB.EXTERNAL_IDS: {TEST_EXTERNAL_IDS_KEY: TEST_EXTERNAL_IDS_VALUE}
    }
    libnmstate.apply({OvsDB.KEY: ovs_config})
    yield ovs_config
    libnmstate.apply(
        {OvsDB.KEY: {OvsDB.EXTERNAL_IDS: {TEST_EXTERNAL_IDS_KEY: None}}}
    )


def test_ovsdb_global_config_untouched_if_not_defined(
    ovsdb_global_config_external_ids,
):
    desired_ovs_config = ovsdb_global_config_external_ids
    libnmstate.apply({})

    current_ovs_config = libnmstate.show()[OvsDB.KEY]
    assert state_match(desired_ovs_config, current_ovs_config)


class TestOvsPatch:
    def test_create_and_remove_patch_port(self):
        patch0_state = {OVSInterface.Patch.PEER: "patch1"}
        patch1_state = {OVSInterface.Patch.PEER: "patch0"}
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PATCH0, patch_state=patch0_state)
        desired_state = bridge.state
        bridge = Bridge(BRIDGE1)
        bridge.add_internal_port(PATCH1, patch_state=patch1_state)
        desired_state[Interface.KEY].extend(bridge.state[Interface.KEY])
        try:
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BRIDGE0)
        assertlib.assert_absent(PATCH0)
        assertlib.assert_absent(PATCH1)

    def test_patch_interface_does_not_have_mtu(self):
        patch0_state = {OVSInterface.Patch.PEER: "patch1"}
        patch1_state = {OVSInterface.Patch.PEER: "patch0"}
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PATCH0, patch_state=patch0_state)
        desired_state = bridge.state
        bridge = Bridge(BRIDGE1)
        bridge.add_internal_port(PATCH1, patch_state=patch1_state)
        desired_state[Interface.KEY].extend(bridge.state[Interface.KEY])
        try:
            libnmstate.apply(desired_state)
            patch0_state = statelib.show_only((PATCH0,))
            assert not patch0_state[Interface.KEY][0].get(Interface.MTU)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

    def test_add_patch_to_existing_interface_invalid(self):
        patch0_state = {OVSInterface.Patch.PEER: "falsepatch"}
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PATCH0)
        desired_state = bridge.state
        bridge = Bridge(BRIDGE1)
        bridge.add_internal_port(PATCH1)
        desired_state[Interface.KEY].extend(bridge.state[Interface.KEY])
        try:
            desired_state[Interface.KEY][1][
                OVSInterface.PATCH_CONFIG_SUBTREE
            ] = patch0_state
            desired_state[Interface.KEY][1][Interface.MTU] = 1500

            with pytest.raises(NmstateValueError):
                libnmstate.apply(desired_state)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
                iface[OVSInterface.PATCH_CONFIG_SUBTREE] = {}
            libnmstate.apply(desired_state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BRIDGE0)
        assertlib.assert_absent(PATCH0)
        assertlib.assert_absent(PATCH1)

    def test_add_patch_to_existing_interface_valid(self):
        patch0_state = {OVSInterface.Patch.PEER: "patch1"}
        patch1_state = {OVSInterface.Patch.PEER: "patch0"}
        bridge = Bridge(BRIDGE0)
        bridge.add_internal_port(PATCH0)
        desired_state = bridge.state
        bridge = Bridge(BRIDGE1)
        bridge.add_internal_port(PATCH1)
        desired_state[Interface.KEY].extend(bridge.state[Interface.KEY])
        try:
            desired_state[Interface.KEY][1].pop(Interface.MTU, None)
            desired_state[Interface.KEY][1].pop(Interface.MAC, None)
            desired_state[Interface.KEY][1][
                OVSInterface.PATCH_CONFIG_SUBTREE
            ] = patch0_state
            desired_state[Interface.KEY][3].pop(Interface.MTU, None)
            desired_state[Interface.KEY][3].pop(Interface.MAC, None)
            desired_state[Interface.KEY][3][
                OVSInterface.PATCH_CONFIG_SUBTREE
            ] = patch1_state

            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)
        finally:
            for iface in desired_state[Interface.KEY]:
                iface[Interface.STATE] = InterfaceState.ABSENT
            libnmstate.apply(desired_state)

        assertlib.assert_absent(BRIDGE1)
        assertlib.assert_absent(BRIDGE0)
        assertlib.assert_absent(PATCH0)
        assertlib.assert_absent(PATCH1)


@pytest.mark.tier1
def test_create_internal_port_with_explict_mtu():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(
        PORT1,
        mac=MAC1,
        mtu=1501,
        ipv4_state={
            InterfaceIPv4.ENABLED: True,
            InterfaceIPv4.ADDRESS: [
                {
                    InterfaceIPv4.ADDRESS_IP: "192.0.2.1",
                    InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                }
            ],
        },
    )

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)


@pytest.mark.tier1
def test_change_mtu_of_existing_internal_port(bridge_with_ports):
    state = statelib.show_only((PORT1,))
    state[Interface.KEY][0][Interface.MTU] = 1501

    libnmstate.apply(state)

    assertlib.assert_state_match(state)


@pytest.mark.tier1
def test_create_ovs_with_internal_ports_in_reverse_order():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT2)
    bridge.add_internal_port(PORT1)

    with bridge.create() as state:
        assertlib.assert_state_match(state)

    assertlib.assert_absent(BRIDGE1)
    assertlib.assert_absent(PORT1)
    assertlib.assert_absent(PORT2)


def test_create_memory_only_ovs_bridge():
    bridge = Bridge(BRIDGE1)
    bridge.add_internal_port(PORT1)

    if nm_major_minor_version() <= 1.28:
        with pytest.raises(NmstateNotSupportedError):
            libnmstate.apply(bridge.state, save_to_disk=False)
    else:
        try:
            libnmstate.apply(bridge.state, save_to_disk=False)
        finally:
            libnmstate.apply(
                {
                    Interface.KEY: [
                        {
                            Interface.NAME: BRIDGE1,
                            Interface.STATE: InterfaceState.ABSENT,
                        }
                    ]
                }
            )


def test_remove_all_ovs_ports(bridge_with_ports):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: BRIDGE1,
                    OVSBridge.CONFIG_SUBTREE: {OVSBridge.PORT_SUBTREE: []},
                }
            ]
        }
    )
    assertlib.assert_absent(PORT1)


def test_modify_state_do_not_remove_unmanaged_ovs(eth1_up):
    with unmanaged_ovs_bridge():
        eth1_up[Interface.KEY][0][Interface.MTU] = 1400
        libnmstate.apply(eth1_up)
        current_state = statelib.show_only((BRIDGE0,))
        assert len(current_state[Interface.KEY])
        assert current_state[Interface.KEY][0][Interface.NAME] == BRIDGE0


@contextmanager
def unmanaged_ovs_bridge():
    rc, _, _ = cmdlib.exec_cmd(
        "ovs-vsctl add-br br0 -- set Bridge br0 fail-mode=secure".split(),
        check=True,
    )
    assert rc == RC_SUCCESS
    try:
        yield
    finally:
        rc, _, _ = cmdlib.exec_cmd("ovs-vsctl del-br br0".split(), check=True)
        assert rc == RC_SUCCESS


def test_expect_failure_when_create_ovs_interface_without_bridge():
    with pytest.raises(NmstateValueError):
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: "ovs0",
                        Interface.TYPE: InterfaceType.OVS_INTERFACE,
                        Interface.STATE: InterfaceState.UP,
                    }
                ]
            }
        )


def test_show_saved_config_with_ovs_bridge_down(bridge_with_ports):
    running_state = statelib.show_only((BRIDGE1,))
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: BRIDGE1,
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only((BRIDGE1,))

    assert saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
    assertlib.assert_state_match_full(saved_state, running_state)


def test_show_saved_config_with_ovs_bridge_iface_same_name(
    bridge_with_same_name_iface,
):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: BRIDGE1,
                    Interface.TYPE: InterfaceType.OVS_BRIDGE,
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only((BRIDGE1,))

    assert len(saved_state[Interface.KEY]) == 2
    assert saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
    assert saved_state[Interface.KEY][1][Interface.STATE] == InterfaceState.UP
    # TODO: assert_state_match_full() does not support same interface.
